#include <socket.h>

#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#ifdef  HAVE_UNISTD_H
# include <unistd.h>
#endif/*HAVE_UNISTD_H*/

#ifdef  HAVE_ARPA_INET_H
# include <arpa/inet.h>
#endif/*HAVE_ARPA_INET_H*/
 
#ifdef  HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif/*HAVE_SYS_SOCKET_H*/

#ifdef  HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif/*HAVE_NETINET_IN_H*/

#ifdef  HAVE_NETDB_H
# include <netdb.h>
#endif/*HAVE_NETDB_H*/

#ifdef HAVE_SYS_TIMES_H
# include <sys/times.h>
#endif/*HAVE_SYS_TIMES_H*/
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif/*TIME_WITH_SYS_TIME*/

#ifdef STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

#if defined( __linux__ )
# include "linux.h"
#else
# include <netinet/ip.h>
# include <netinet/ip_icmp.h>
#endif /*defined(__linux__)*/

#define  MAXPACKET   65535
#define  PKTSIZE     64 
#define  HDRLEN      ICMP_MINLEN
#define  DATALEN     (PKTSIZE-HDRLEN)
#define  MAXDATA     (MAXPKT-HDRLEN-TIMLEN)

struct SOCKET_T
{ 
  int      socket;
  int      conn;
  char     *hostname;
  int      port;
  int      timeout;
  struct   hostent     *hp; 
  struct   protoent    *proto;
  char     proto_buf[1024];
  struct   sockaddr_in addr;
  struct   icmp  *icp;
  int      domain;
  int      type;
  BOOLEAN  connected;
  BOOLEAN  debug;
  PROTOCOL protocol;
  char     *protoname;
  struct   timeval mytime;
  int      rrt;
};  

private BOOLEAN __is_ipaddr(const char *addr);
private BOOLEAN __setup(SOCKET this);
private int     __socket_block(SOCKET this, BOOLEAN block);
private BOOLEAN __tcp_socket_check(SOCKET this);
private BOOLEAN __icmp_socket_check(SOCKET this);
private BOOLEAN __socket_sendto(SOCKET this);
private BOOLEAN __socket_recvfrom(SOCKET this);
private char *  __rlookup(struct  in_addr in);
private int     __in_checksum(u_short *buf, int len);
private int     __elapsed_time(struct timeval *starttime);
  
SOCKET
new_socket(char *hostname, int port, PROTOCOL protocol)
{
  SOCKET this;

  this = xcalloc(sizeof(*this),1);
  this->hostname  = xstrdup(hostname);
  this->port      = port;
  this->timeout   = 15;
  this->connected = FALSE;
  this->debug     = FALSE;
  this->protocol  = protocol;
  switch(this->protocol){
    case ICMP:
      this->protoname = xstrdup("icmp");
      this->domain    = AF_INET;
      this->type      = SOCK_RAW;
      break;
    case TCP:
    default:
      this->protoname = xstrdup("tcp");
      this->domain    = AF_INET;
      this->type      = SOCK_STREAM;
      break;
  }
  __setup(this);
  return this;
}

SOCKET
socket_destroy(SOCKET this){
  socket_close(this);
  xfree(this->hostname);
  xfree(this);
  this = NULL;
  return this; 
}

BOOLEAN 
socket_close(SOCKET this){
  int ret;

  if(this->socket != -1){
    if((ret = close(this->socket)) < 0)
      debug(this->debug, "unable to close the socket %s:%d",    __FILE__, __LINE__);
    this->socket = -1;
    this->connected = FALSE;
  }
  return TRUE;
}

void
socket_set_port(SOCKET this, int port){
  this->port = port;
  return;
}

void 
socket_set_timeout(SOCKET this, int timeout){
  this->timeout = timeout;
  return;
}

int 
socket_get_port(SOCKET this){
  return this->port;
}

int
socket_get_timeout(SOCKET this){
  return this->timeout;
}

void
socket_set_debug(SOCKET this, BOOLEAN debug){
  this->debug = debug;
}

char *
socket_get_hostname(SOCKET this){
  return this->hostname;
}

char *
socket_get_ptr(SOCKET this){
  return __rlookup(this->addr.sin_addr);
}

BOOLEAN
socket_check(SOCKET this){
  if(this->protocol==ICMP){
    return __icmp_socket_check(this); 
  } else {
    return __tcp_socket_check(this); 
  }
}

private BOOLEAN
__tcp_socket_check(SOCKET this){
  int res;
  struct sockaddr_in cli;

  this->addr.sin_port = htons(this->port);
  this->addr.sin_family = AF_INET;
  this->socket = socket(AF_INET,SOCK_STREAM,0);
  if(this->socket < 0) {
    debug(this->debug, "socket failure");
    return FALSE;
  }

  if((__socket_block(this, FALSE)) < 0){
    debug(this->debug, "socket: unable to set socket to non-blocking %s:%d", __FILE__, __LINE__);
    return FALSE;
  }

  this->conn = connect(this->socket,(struct sockaddr *)&this->addr,sizeof(this->addr));
  if(this->conn < 0 && errno != EINPROGRESS){
    socket_close(this); 
    return FALSE;
  } else {
    struct timeval timeout;
    fd_set rs;
    fd_set ws;
    FD_ZERO(&rs);
    FD_ZERO(&ws);
    FD_SET(this->socket, &rs);
    FD_SET(this->socket, &ws);
    memset((void *)&timeout, '\0', sizeof(struct timeval));
    timeout.tv_sec  = socket_get_timeout(this);
    timeout.tv_usec = 0;
    res = select(this->socket+1, &rs, &ws, NULL, &timeout);
    if(res < 1){
      debug(this->debug, "socket: connection timed out\n");
      socket_close(this);
      return FALSE;
    } else {
      /**
       * If we reconnect and receive EISCONN, then we have a successful connection
       */
      res = connect(this->socket, (struct sockaddr *)&cli, sizeof(struct sockaddr_in));
      if((res < 0)&&(errno != EISCONN)){
        debug(this->debug, "socket: unable to connect %s:%d", __FILE__, __LINE__);
        socket_close(this);
        return FALSE;
      }
      this->connected = TRUE;
    } 
  }  
  if(this->connected == TRUE){
    socket_close(this);
    return TRUE;
  } else {
    socket_close(this);
    return FALSE;
  }
}

private BOOLEAN 
__icmp_socket_check(SOCKET this){
  BOOLEAN ret;

   (void) gettimeofday(&this->mytime, (struct timezone *)NULL);
  if((ret = __socket_sendto(this)) == FALSE){
    socket_close(this);
    return ret;
  }

  do {
    this->rrt = __elapsed_time(&this->mytime);
    if (this->rrt > this->timeout * 1000) {
      socket_close(this);
      return FALSE;
    }
  } while(__socket_recvfrom(this));
  socket_close(this);

  if(this->rrt < 0) return FALSE;
  return TRUE;
}

private BOOLEAN
__socket_sendto(SOCKET this){
  int ss;
  int len = HDRLEN+DATALEN;
  unsigned char buf[len];

  this->icp = (struct icmp *)buf;
  this->icp->icmp_type  = ICMP_ECHO;
  this->icp->icmp_code  = 0;
  this->icp->icmp_cksum = 0;
  this->icp->icmp_id    = getpid() & 0xFFFF;
  this->icp->icmp_cksum = __in_checksum((u_short *)this->icp, len);

  if((ss = sendto(this->socket, buf, sizeof(buf), 0, (struct sockaddr *)&this->addr, sizeof(this->addr))) < 0){
    return FALSE;
  }
  if(ss != len){
    return FALSE;
  }

  return TRUE;
}

private BOOLEAN
__socket_recvfrom(SOCKET this){
  int res;
  int from;
  int len = HDRLEN+DATALEN; 
  unsigned char buf[len];
  struct timeval to;
  fd_set readset;
  struct icmp        *icp;
  struct sockaddr_in faddr;

  to.tv_sec  = this->timeout/100000;
  to.tv_usec = (this->timeout - (to.tv_sec * 100000)) * 10;

  FD_ZERO(&readset);
  FD_SET(this->socket, &readset);

  /* we use select to see if there is any activity
     on the socket.  If not, then we've requested an
     unreachable network and we'll time out here. */
  if((res = select(this->socket + 1, &readset, NULL, NULL, &to)) < 1){
    this->rrt = -4;
    return FALSE;
  }

  from = sizeof faddr;

  if((res = recvfrom(this->socket, buf, len, 0, (struct sockaddr*)&faddr, &from)) < 0){
    this->rrt = -5;
    perror("recvfrom");
    return FALSE;
  } 

  icp = (struct icmp *)(buf + len);
  if(faddr.sin_addr.s_addr != this->addr.sin_addr.s_addr){
    return TRUE;
  }

  return FALSE;
}


private char *
__rlookup(struct  in_addr in)
{
  struct  hostent *hp;
  static  char    buf[512];

  hp = gethostbyaddr((char *)&in,sizeof(in),AF_INET);
  if(hp == NULL){
    return(inet_ntoa(in));
  } else {
    snprintf(buf,sizeof(buf), "%s (%s)",hp->h_name,inet_ntoa(in));
    return(buf);
  }
}

private BOOLEAN
__is_ipaddr(const char *addr){
  int i;
  char ipc[4][4], ignore;

  int ret = sscanf(
    addr,
    "%3[0123456789]." 
    "%3[0123456789]." 
    "%3[0123456789]." 
    "%3[0123456789]%c",
    ipc[0], ipc[1], ipc[2], ipc[3], &ignore
  );

  if (ret != 4) return FALSE;

  i = atoi(ipc[0]); if (i < 0 || i > 255) return FALSE;
  i = atoi(ipc[1]); if (i < 0 || i > 255) return FALSE;
  i = atoi(ipc[2]); if (i < 0 || i > 255) return FALSE;
  i = atoi(ipc[3]); if (i < 0 || i > 255) return FALSE;

  return TRUE; 
}

private BOOLEAN
__setup(SOCKET this){
  this->proto  = getprotobyname(this->protoname); 
  if(this->proto==NULL){
    return FALSE;
  }

  if(__is_ipaddr(this->hostname)){
    long i;
    if((i = inet_addr(this->hostname)) == -1L) {
      debug(this->debug,"%s: address unparsable\n",this->hostname);
      return FALSE;
    }
    (void)bcopy((char *)&i,(char *)&this->addr.sin_addr,sizeof(i));
  } else {
    unsigned short last; 
    if((this->hp = gethostbyname(this->hostname)) == (struct hostent *)0) {
      debug(this->debug,"%s: address unknown\n",this->hostname);
      return FALSE;
    }
    (void)bcopy(this->hp->h_addr,(char *)&this->addr.sin_addr,this->hp->h_length);
    last = ntohl(this->addr.sin_addr.s_addr) & 0xFF;
    if(( last == 0x00 ) || ( last == 0xFF )){
      return FALSE;
    }
  }

  if((this->socket = socket(this->domain, this->type, this->proto->p_proto)) < 0){
    switch(errno){
      case EPROTONOSUPPORT: { debug(this->debug, "unsupported protocol %s:%d",  __FILE__, __LINE__); break; }
      case EMFILE:          { debug(this->debug, "descriptor table full %s:%d", __FILE__, __LINE__); break; }
      case ENFILE:          { debug(this->debug, "file table full %s:%d",       __FILE__, __LINE__); break; }
      case EACCES:          { debug(this->debug, "permission denied %s:%d",     __FILE__, __LINE__); break; }
      case ENOBUFS:         { debug(this->debug, "insufficient buffer %s:%d",   __FILE__, __LINE__); break; }
      default:              { debug(this->debug, "unknown socket error %s:%d",  __FILE__, __LINE__); break; }
    } socket_close(this); return FALSE;
  }

  return TRUE;
}

private int
__in_checksum(u_short *buf, int len) {
  register long sum = 0;
  u_short  answer = 0;

  while(len > 1){
    sum += *buf++;
    len -= 2;
  }

  if(len == 1){
    *(u_char*)(&answer) = *(u_char*)buf;
    sum += answer;
  }
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);     
  answer = ~sum;     

  return answer;
}

/**
 * set socket to non-blocking (FALSE) or blocking (TRUE)
 */
private int
__socket_block(SOCKET this, BOOLEAN block)
{
#if HAVE_FCNTL_H 
  int flags;
  int retval;
#elif defined(FIONBIO)
  ioctl_t status;
#else
  return this->socket;
#endif

  if(this->socket==-1){
    return this->socket;
  }

#if HAVE_FCNTL_H 
  if((flags = fcntl(this->socket, F_GETFL, 0)) < 0){
    switch(errno){
      case EACCES: { debug(this->debug, "EACCES %s:%d",                 __FILE__, __LINE__); break; }
      case EBADF:  { debug(this->debug, "bad file descriptor %s:%d",    __FILE__, __LINE__); break; }
      case EAGAIN: { debug(this->debug, "address is unavailable %s:%d", __FILE__, __LINE__); break; }
      default:     { debug(this->debug, "unknown network error %s:%d",  __FILE__, __LINE__); break; }
    } return -1;
  }

  if(block){
    flags &= ~O_NDELAY;
  } else {
    flags |=  O_NDELAY;
    #if (defined(hpux) || defined(__hpux) || defined(__osf__)) || defined(__sun)
    #else
    flags |=  O_NONBLOCK;
    #endif
  }

  if((retval = fcntl(this->socket, F_SETFL, flags)) < 0){
    debug(this->debug, "unable to set fcntl flags %s:%d", __FILE__, __LINE__);
    return -1;
  }
  return retval;

#elif defined(FIONBIO)
  status = block ? 0 : 1;
  return ioctl(this->socket, FIONBIO, &status);
#endif
}


private int
__elapsed_time(struct timeval *starttime){
  struct timeval *newtime;
  int elapsed;
  newtime = (struct timeval*)malloc( sizeof(struct timeval));
  gettimeofday(newtime,NULL);
  elapsed = 0;

  if((newtime->tv_usec - starttime->tv_usec) > 0){
    elapsed += (newtime->tv_usec - starttime->tv_usec)/1000 ;
  } else {
    elapsed += (1000000 + newtime->tv_usec - starttime->tv_usec) /1000;
    newtime->tv_sec--;
  }
  if((newtime->tv_sec - starttime->tv_sec ) > 0 ){
    elapsed += 1000 * (newtime->tv_sec - starttime->tv_sec);
  }
  if(elapsed < 1)
    elapsed = 1;

  free(newtime);
  return elapsed;
}
